/*
 * MacSamba
 * <http://ghostwood.org/software/macSamba.html>
 * Copyright 2006-2007 Nick Howes <nick@ghostwood.org>
 * tabs every 4 spaces
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; version 2
 * of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <math.h>
#include <string.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <IOKit/usb/IOUSBLib.h>

#include "util.h"
#include "devices.h"
#include "usb.h"
#include "ops.h"

#define GREET_INFO "Welcome to MacSamba 1.3  2006-2009 Nick Howes\n"
#define USAGE_INFO "MacSamba [-f FILE] [-b ADDR] [-e] [-m] [-r]\n"
#define OPT_INFO_1 "  -f FILE    a raw binary data file to write to flash\n"
#define OPT_INFO_2 "  -b ADDR    an address offset (in the flash space)\n"
#define OPT_INFO_3 "  -e         do a bulk erase before writing (otherwise do automatic page erases while writing)\n"
#define OPT_INFO_4 "  -m         re-map the flash space to boot from flash\n"
#define OPT_INFO_5 "  -s         set the security bit\n"
#define OPT_INFO_6 "  -r         reset the CPU after completing other tasks\n"

#define PMC_MCKR_ADDR			0xFFFFFC30	// PMC Master Clock Register
#define PMC_MCKR_MASK			0x1F // PRES and CSS fields
#define PMC_MCKR_DATA			0x01 // PRES=0 : no prescaling; CSS=1 : Main Clock Selected 18.432MHz

#define MC_FMR_ADDR				0xFFFFFF60	// Flash Mode Register
#define MC_FMR_DATA_NVM			0x00130080	// FMCN=19 (1.0us), NEBP=1
#define MC_FMR_DATA_FLASH		0x001C0080	// FMCN=28 (1.5us), NEBP=1 (no page erase)
#define MC_FMR_DATA_FLASH_E		0x001C0000	// FMCN=28 (1.5us), NEBP=0 (page erase)

#define MC_FCR_ADDR				0xFFFFFF64	// Flash Command Register
#define MC_FCR_CMD_ERASE_ALL	0x5A000008
#define MC_FCR_CMD_WRITE_PAGE	0x5A000001 // | (page << 8)
#define MC_FCR_CMD_SET_SECURITY	0x5A00000F //
#define MC_FCR_CMD_SET_NVM2		0x5A00020B // boot from flash  NOTE: change FMR FMCN for this command

#define MC_FSR_ADDR				0xFFFFFF68	// Flash Status Register


//this is painfully slow
void dumpDataBlock(usb_t usb, unsigned int addr, int blocks) {
	int block, line, col;
	unsigned int data;
	
	for (block=0 ; block<blocks ; block++) {
		for (line=0 ; line<16 ; line++) {
			printf("%8X : ", addr);
			for (col=0 ; col<16 ; col++) {
				readByte(usb, addr++, &data);
				printf("%2X ", data);
			}
			printf("\n");
		}
		printf("\n");
	}
	printf("\n");
}


void doComms(int fd, int bulkErase, int baseAddr, int remap, int dump, int security, int reset) {
	usb_t usb;
	unsigned int addr, data, readLen, page;
	deviceInfo_t* dev;
	char buffer[FLASH_PAGE_SIZE];
	int dotCounter = 18;

	usb = getUSB();
	
	dev = getDeviceInfo(usb);
	if (dev == NULL) failExit("Unknown CPU\n", 0);
	printf("Device is %s\n", dev->name);
	

	//check that Master Clock is as expected (Main Oscillator which is 18.432MHz)
	readWord(usb, PMC_MCKR_ADDR, &data);
	if ((data & PMC_MCKR_MASK) != PMC_MCKR_DATA) failExit("Unexpected PMC_MCKR setting", data);
	
	
	//erase flash
	if (bulkErase) {
	
		//Set Flash Mode Register for Flash Programming (no page auto erase)
		writeWord(usb, MC_FMR_ADDR, MC_FMR_DATA_FLASH);
		readWord(usb, MC_FMR_ADDR, &data);
		if (data != MC_FMR_DATA_FLASH) failExit("Unexpected MC_FMR setting", data);

		printf("Erasing Flash ...");
		waitFlashReady(usb);
		writeWord(usb, MC_FCR_ADDR, MC_FCR_CMD_ERASE_ALL);
		checkFlashProgOK(usb);
		printf(" OK\n");
		
	} else {
	
		//Set Flash Mode Register for Flash Programming (with page auto erase)
		printf("Using Automatic Page Erase\n");
		writeWord(usb, MC_FMR_ADDR, MC_FMR_DATA_FLASH_E);
		readWord(usb, MC_FMR_ADDR, &data);
		if (data != MC_FMR_DATA_FLASH_E) failExit("Unexpected MC_FMR setting", data);

	}
	
	//write data, etc.
	if (fd >= 0) {
		if (baseAddr != 0) 	printf("Starting Address is 0x%x\n", baseAddr);
		printf("Programming Flash ");
		addr = baseAddr + dev->flashAddr;
		for (;;) {
		
			readLen = read(fd, buffer, sizeof(buffer));
			if (readLen == 0) break;
			if (readLen < 0) failExit("file read failed\n", 0);
			
			writeBlock(usb, addr, buffer, readLen);
	
			page = (addr & 0xfffff) / FLASH_PAGE_SIZE;
			if (page >= (dev->flashPageCount)) failExit("not enough flash space\n", 0);
			
			waitFlashReady(usb);
			writeWord(usb, MC_FCR_ADDR, MC_FCR_CMD_WRITE_PAGE | (page << 8));
			checkFlashProgOK(usb);
	
			addr += readLen;
	
			printf(".");
			if (++dotCounter > 80) {
				printf("\n");
				dotCounter = 0;
			}
			fflush(stdout);
	
		}
		printf(" OK\n");
	}
	
	//dump data
	if (dump) dumpDataBlock(usb, dev->flashAddr, dev->flashPageCount);
	

	//Set Flash Mode Register for NVM etc. Programming
	writeWord(usb, MC_FMR_ADDR, MC_FMR_DATA_NVM);
	readWord(usb, MC_FMR_ADDR, &data);
	if (data != MC_FMR_DATA_NVM) failExit("Unexpected MC_FMR setting", data);


	//set GPNVM bit 2 to remap flash
	if (remap) {
		printf("Remapping Flash ...");
		waitFlashReady(usb);
		writeWord(usb, MC_FCR_ADDR, MC_FCR_CMD_SET_NVM2);
		checkFlashProgOK(usb);
		printf(" OK\n");
	}

	//set security bit
	if (security) {
		printf("Setting Security bit ...");
		waitFlashReady(usb);
		writeWord(usb, MC_FCR_ADDR, MC_FCR_CMD_SET_SECURITY);
		checkFlashProgOK(usb);
		printf(" OK\n");
	}

	if (reset) {
		printf("Resetting target ...");
		resetTarget(usb);
		printf(" OK\n");
	}
	
	//clean up
	(*(usb.interface))->USBInterfaceClose(usb.interface);
	(*(usb.interface))->Release(usb.interface);
	
}


int main(int argc, char **argv) {
	int fd = -1;
	int bulkErase = 0;
	int baseAddress = 0;
	int remap = 0;
	int dump = 0;
	int reset = 0;
	int security = 0;
	
	extern char *optarg;
	extern int optind;
	int ch;
	
	printf(GREET_INFO);

	//handle user input
	while ((ch = getopt(argc, argv, "emrdf:b:")) != -1) {
		switch (ch) {
			case 'e':
				bulkErase = 1;
				break;
			case 'm':
				remap = 1;
				break;
			case 'r':
				reset = 1;
				break;
			case 's':
				security = 1;
				break;
			case 'd':
				dump = 1;
				break;
			case 'f':
				fd = open(optarg, O_RDONLY);
				if (fd < 0) failExit("cannot open specified file", 0);
				break;
			case 'b':
				sscanf(optarg, "%x", &baseAddress);
				break;
			default:
				printf (USAGE_INFO);
				printf (OPT_INFO_1);
				printf (OPT_INFO_2);
				printf (OPT_INFO_3);
				printf (OPT_INFO_4);
				printf (OPT_INFO_5);
				printf (OPT_INFO_6);
				exit(2);
		}
	}
	
	doComms(fd, bulkErase, baseAddress, remap, dump, security, reset);
	
	//clean up
	close(fd);
	exit(0);
}
