/*
 * MacSamba
 * <http://ghostwood.org/software/macSamba.html>
 * Copyright 2006 Nick Howes <nick@ghostwood.org>
 * tabs every 4 spaces
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; version 2
 * of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */


#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>
#include <IOKit/usb/IOUSBLib.h>

#include "devices.h"
#include "usb.h"
#include "util.h"

#define DBGU_ID_ADDR	0xFFFFF240	// Chip ID code

#define MC_FSR_ADDR		0xFFFFFF68	// Flash Status Register
#define MC_FSR_PROGE	0x8 // PROGE bit

#define RSTC_CR_ADDR			0xFFFFFD00	// Reset Controller Control Register
#define RSTC_CR_DATA_RST_ALL	0xA500000D	// PROCRST=1, RERRST=1, EXTRST=1


#define SLEEP_TIME 10000
#define RETRIES 100

int readWord(usb_t usb, unsigned int addr, unsigned int* word) {
	char buf[256], *p;
	int n, r;

	n = snprintf(buf, sizeof(buf), "w%X,#", addr);
	writeUSB(usb, buf, n);
	n = readUSB(usb, buf, sizeof(buf));
	p = buf;
	while (isspace(*p)) p++; //skip any whitespace
	r = sscanf(p, "0x%x", word);
	if (r == 1) {
		return 0;
	} else {
		return -1;
	}
}

int readByte(usb_t usb, unsigned int addr, unsigned int* word) {
	char buf[256], *p;
	int n, r;

	n = snprintf(buf, sizeof(buf), "o%X,#", addr);
	writeUSB(usb, buf, n);
	n = readUSB(usb, buf, sizeof(buf));
	p = buf;
	while (isspace(*p)) p++; //skip any whitespace
	r = sscanf(p, "0x%x", word);
	if (r == 1) {
		return 0;
	} else {
		return -1;
	}
}

int writeWord(usb_t usb, unsigned int addr, unsigned int word) {
	char buf[256];
	int n;

	n = snprintf(buf, sizeof(buf), "W%X,%X#", addr, word);
	writeUSB(usb, buf, n);
	n = readUSB(usb, buf, sizeof(buf));
	return 0;
}

//TODO: the "S" transaction doesn't work for the flash buffer - perhaps because it doesn't use 32-bit transactions internally
/*
int writeBlock(usb_t usb, unsigned int addr, char* buffer, unsigned int size) {
	char buf[256];
	int n;

	n = snprintf(buf, sizeof(buf), "S%X,%X#", addr, size);
	writeUSB(usb, buf, n); //command
	readCountUSB(usb, 2); //wait for CRLF //TODO: check actual response?
	writeUSB(usb, buffer, size);//data
	n = readUSB(usb, buf, sizeof(buf));
	return 0;
}
*/
int writeBlock(usb_t usb, unsigned int addr, char* buffer, unsigned int size) {
	int i;
	unsigned int data;
	
	for (i=0 ; i<size ; i+=4) {
		data = (*buffer) & 0xff;
		++buffer;
		data |= ((*buffer) & 0xff) << 8;
		++buffer;
		data |= ((*buffer) & 0xff) << 16;
		++buffer;
		data |= ((*buffer) & 0xff) << 24;
		++buffer;
		writeWord(usb, addr + i, data);
	}

	return 0;
}

deviceInfo_t* getDeviceInfo(usb_t usb) {
	unsigned int data;

	readWord(usb, DBGU_ID_ADDR, &data);
	return getDeviceByID(data);
}

void waitFlashReady(usb_t usb) {
	int retry = 0;
	unsigned int data;

	do {
		readWord(usb, MC_FSR_ADDR, &data);
		usleep(SLEEP_TIME);
		if ((data & 1) == 1) return;
	} while (++retry < RETRIES);
	
	failExit("timeout waiting for flash to become ready", 0);
}

void checkFlashProgOK(usb_t usb) {
	unsigned int data;

	waitFlashReady(usb);
	readWord(usb, MC_FSR_ADDR, &data);
	if (data & MC_FSR_PROGE) failExit("Flash Programming Error", data);
}


void resetTarget(usb_t usb) {
	char buf[256];
	int n;

	n = snprintf(buf, sizeof(buf), "W%X,%X#", RSTC_CR_ADDR, RSTC_CR_DATA_RST_ALL);
	writeUSB(usb, buf, n);
}


